! This program reads simulation output data from the Finnish hybrid web 
! archive and prepares files with electric and magnetic fields. 
!
! The cheese inputfile, i.e. inputcheese.m, is read and so are the output 
! files from the hybrid simulations. 
! A grid is created based on the parameters in inputcheese.m, and the 
! hybrid simulation files must contain data on that same grid. 


program preparefields

  use SpecificTypes
  use ieee_arithmetic

  implicit none

! input variables
  type (InputGeneral) inpar

! grid
  double precision Lx, Ly, Lz, dxyz(3)
  double precision, allocatable :: X(:), Y(:), Z(:)
  type (GridLevel), allocatable :: grid(:)

! fields
  double precision, allocatable :: E(:,:,:,:), B(:,:,:,:), cur(:,:,:,:), &
       ue(:,:,:,:), dens(:,:,:)
  double precision B1(3), v1(3)

! particle species
  type(species), allocatable :: particle(:)
  type(species) CurAndMag

! Universal constants
  double precision elc

! Other variables
  integer Nretries, attempts, gl, ii, jj, kk, ll, mm, ierr
  character fieldname*7, snumber*2, glnumber*2


! Initialise a couple of things
  parameter(Nretries = 10)
  parameter(elc = 1.6021765300d-19)

! Read input data
  attempts = 0
  call GetGeneralInput(inpar, attempts, Nretries)   

  allocate(grid(inpar%Ngridlevels))
  attempts = 0
  call GetGridLevels('', inpar%Ngridlevels, grid, attempts, Nretries)

! Loop over all the grid levels
  do gl = 1, inpar%Ngridlevels
     write(glnumber,fmt='(i2.2)') gl

! System size
     Lx = grid(gl)%xmax - grid(gl)%xmin
     Ly = grid(gl)%ymax - grid(gl)%ymin
     Lz = grid(gl)%zmax - grid(gl)%zmin

! Set grid cell size
     dxyz(1) = Lx/(grid(gl)%Nx-1)
     dxyz(2) = Ly/(grid(gl)%Ny-1)
     dxyz(3) = Lz/(grid(gl)%Nz-1)

! Allocate and create x, y, and z vectors for the grid
     allocate(X(grid(gl)%Nx))
     allocate(Y(grid(gl)%Ny))
     allocate(Z(grid(gl)%Nz))
     do ii = 1, grid(gl)%Nx
        X(ii) = grid(gl)%xmin + dxyz(1)*dble(ii-1)
     end do
     do ii = 1, grid(gl)%Ny
        Y(ii) = grid(gl)%ymin + dxyz(2)*dble(ii-1)
     end do
     do ii = 1, grid(gl)%Nz
        Z(ii) = grid(gl)%zmin + dxyz(3)*dble(ii-1)
     end do

! Allocate fields and initialise
     allocate(B(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))  ! Magnetic field
     allocate(E(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))  ! Electric field
     allocate(cur(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))! Current
     allocate(ue(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz)) ! Electron bulk vel.
     allocate(dens(grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz)) ! Electron number dens
     B    = 0.0d0
     E    = 0.0d0
     cur  = 0.0d0

! Data specific to each species
     allocate(particle(inpar%NhybSpecies))
! Get species information from the input file, i.e. the HWA file names.
     attempts = 0
     call GetSpecies('', inpar%NhybSpecies, particle, attempts, Nretries)
! Allocate data variables for each species and get the data.
     do ii = 1, inpar%NhybSpecies
        allocate(particle(ii)%B(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
        allocate(particle(ii)%vel(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
        allocate(particle(ii)%cur(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
        allocate(particle(ii)%rho(grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
        allocate(particle(ii)%dens(grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
        particle(ii)%B    = 0.0d0   ! B field
        particle(ii)%vel  = 0.0d0   ! velocity
        particle(ii)%cur  = 0.0d0   ! current density
        particle(ii)%rho  = 0.0d0   ! mass density
        particle(ii)%dens = 0.0d0   ! number density
        particle(ii)%hybfilename = trim(particle(ii)%hybfilename)//glnumber
        attempts = 0
        call GetHybridData('', grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, X, Y, Z, &
             dxyz, grid(gl)%xmin, grid(gl)%ymin, grid(gl)%zmin, particle(ii), &
             attempts, Nretries, ierr)
     end do

! Assemble the fields
  ! We use the variable CurAndMag of the particle type for B field and current
  ! so that we can read it with the GetHybridData subroutine. If reading those 
  ! quantities from their own files fails, we take what is in the first 
  ! particle species file. If that fails too, the values shall be zero.
     allocate(CurAndMag%B(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
     allocate(CurAndMag%cur(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
     CurAndMag%B    = 0.0d0   ! B field
     CurAndMag%cur  = 0.0d0   ! current density
  ! get magnetic field
     CurAndMag%hybfilename = trim(inpar%magfilename)//glnumber
     attempts = 0
     call GetHybridData('', grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, X, Y, Z, &
          dxyz, grid(gl)%xmin, grid(gl)%ymin, grid(gl)%zmin, CurAndMag, &
          attempts, Nretries, ierr)
     if (ierr==0) then
        B = CurAndMag%B
        particle(1)%B = B
     else
        B = particle(1)%B
        do ii = 2, inpar%NhybSpecies
           write (*,*) 'Check B, should be zero: ', &
                sum((particle(ii)%B - B)**2.0d0)
        end do
     end if
     ! get current
     CurAndMag%hybfilename = trim(inpar%curfilename)//glnumber
     attempts = 0
     call GetHybridData('', grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, X, Y, Z, &
          dxyz, grid(gl)%xmin, grid(gl)%ymin, grid(gl)%zmin, CurAndMag, &
          attempts, Nretries, ierr)
     if (ierr==0) then
        cur = CurAndMag%cur
        particle(1)%cur = cur
     else
        cur = particle(1)%cur
        do ii = 2, inpar%NhybSpecies
           write (*,*) 'Check J, should be zero: ', &
                sum((particle(ii)%cur - cur)**2.0d0)
        end do
     end if

     write (*,*) 'Level ', glnumber, &
          ': sum(B**2.0d0), should not be zero: ', sum(B**2.0d0)
     write (*,*) 'Level ', glnumber, &
          ': sum(J**2.0d0), should not be zero: ', sum(cur**2.0d0)

! Total density
     ue   = 0.0d0
     dens = 0.0d0
     do ii = 1, inpar%NhybSpecies
        dens = dens + particle(ii)%dens * particle(ii)%chargestate
     end do

!       j = -e*ne*ue + e*n(H+)*U(H+) + e*n(H2O+)*U(H2O+)
!  =>  ue = (n(H+)*U(H+) + n(H2O+)*U(H2O+))/ne - j/(e*ne)
! and we assume quasi-neutrality, i.e.  ne = n(H+) + n(H2O+)
     do ii = 1, inpar%NhybSpecies
        do jj = 1, 3
           ue(jj,:,:,:) = ue(jj,:,:,:) + &
                particle(ii)%dens * particle(ii)%chargestate * &
                particle(ii)%vel(jj,:,:,:)/dens
        end do
     end do
     do jj = 1, 3
        ue(jj,:,:,:) = ue(jj,:,:,:) - cur(jj,:,:,:)/(dens*elc)
     end do
! Finally we have E = -ue x B = B x ue
     do mm = 1, grid(gl)%Nz
        do ll = 1, grid(gl)%Ny
           do kk = 1, grid(gl)%Nx
              B1 = B(:,kk,ll,mm)
              v1 = ue(:,kk,ll,mm)
              E(1,kk,ll,mm) = B1(2)*v1(3) - B1(3)*v1(2)
              E(2,kk,ll,mm) = B1(3)*v1(1) - B1(1)*v1(3)
              E(3,kk,ll,mm) = B1(1)*v1(2) - B1(2)*v1(1)
           end do
        end do
     end do

! Go through it all and replace NaNs by zeros
     do mm = 1, grid(gl)%Nz
        do ll = 1, grid(gl)%Ny
           do kk = 1, grid(gl)%Nx
              do jj = 1, 3
                 if (ieee_is_nan(E(jj,kk,ll,mm))) then
                    write (*,*) 'E(',jj,',',kk,',',ll,',',mm,')=', &
                         E(jj,kk,ll,mm)
                    E(jj,kk,ll,mm) = 0.0d0
                 end if
              end do
           end do
        end do
     end do

! Now we should be ready to write the files

! Binary field dump
     attempts = 0
     call DumpFieldDump(B, E, gl, grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, &
          attempts, Nretries)

! b-field in text file
     ii = 1
     fieldname = 'Bfi_'//glnumber//'_'
     attempts = 0
     call DumpVectorField(B, grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, &
          'outp/datfiles/Bfield/', fieldname, &
          particle(ii)%hybiteration, attempts, Nretries)

! E-field in text file
     ii = 1
     fieldname = 'Efi_'//glnumber//'_'
     attempts = 0
     call DumpVectorField(E,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz, &
          'outp/datfiles/Efield/',fieldname, &
          particle(ii)%hybiteration, attempts, Nretries)

! Current in text file
     ii = 1
     attempts = 0
     fieldname = 'cur_'//glnumber//'_'
     call DumpVectorField(particle(ii)%cur,grid(gl)%Nx,grid(gl)%Ny, &
          grid(gl)%Nz, 'outp/datfiles/current/', &
          fieldname, particle(ii)%hybiteration, attempts, Nretries)

! Densities in text file
     do ii = 1, inpar%NhybSpecies
        attempts = 0
        write(snumber,fmt='(i2.2)') ii
        fieldname = 'n'//snumber//'_'//glnumber//'_'
        call DumpScalarField(particle(ii)%dens,grid(gl)%Nx,grid(gl)%Ny, &
             grid(gl)%Nz, 'outp/datfiles/density/', &
             fieldname, particle(ii)%hybiteration, attempts, Nretries)
     end do

! Deallocate before starting over
     deallocate(X)
     deallocate(Y)
     deallocate(Z)
     deallocate(B)
     deallocate(E)
     deallocate(cur)
     deallocate(ue)
     deallocate(dens)
     deallocate(particle)
     deallocate(CurAndMag%B)
     deallocate(CurAndMag%cur)


  end do

end program preparefields

