! This program computes trajectories and collects statistics, which it then 
! saves.

program cheddar

  use SpecificTypes

  implicit none

  type (InputGeneral) inpar

  integer ii, jj, kk, gl, nextno
  
  double precision B0(3)

  double precision dt, rr, pi

  double precision E0(3), Eperp(3), dxyz(3), Lx, Ly, Lz
  type (GridLevel), allocatable :: grid(:)

  double precision, allocatable :: dens(:,:,:)

  double precision, allocatable :: position_init(:,:), velocity_init(:,:)
	   
  double precision, allocatable :: position(:,:), velocity(:,:)

  double precision E1(3), E1par(3), E1perp(3), E1mag, rndph, fluctphase

  integer Nretries, attempts

  type(fprobe), allocatable :: probe(:)
  character probefile*5, probeno*2

! General constants
  parameter(pi = 3.14159265358979323846264338327950288419716d0)
  parameter(Nretries = 10)

! Read input data
  attempts = 0
  call GetGeneralInput(inpar, attempts, Nretries)

  allocate(grid(inpar%Ngridlevels))
  attempts = 0
  call GetGridLevels('', inpar%Ngridlevels, grid, attempts, Nretries)
  do gl = 1, inpar%Ngridlevels
     ! Grid size
     grid(gl)%Lx = grid(gl)%xmax - grid(gl)%xmin
     grid(gl)%Ly = grid(gl)%ymax - grid(gl)%ymin
     grid(gl)%Lz = grid(gl)%zmax - grid(gl)%zmin
     ! Grid cell size
     grid(gl)%dxyz(1) = grid(gl)%Lx/(grid(gl)%Nx-1)
     grid(gl)%dxyz(2) = grid(gl)%Ly/(grid(gl)%Ny-1)
     grid(gl)%dxyz(3) = grid(gl)%Lz/(grid(gl)%Nz-1)
     ! Field allocation
     allocate(grid(gl)%E(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
     allocate(grid(gl)%B(3,grid(gl)%Nx,grid(gl)%Ny,grid(gl)%Nz))
     ! Field loading
     attempts = 0
     call LoadFieldDump(grid(gl)%B, grid(gl)%E, gl, &
          grid(gl)%Nx, grid(gl)%Ny, grid(gl)%Nz, attempts, Nretries)
  end do

 ! Initialise timestep variable to the value in the input file
  dt           = inpar%dt ! dt

! Allocate variables for initial position and velocity, and for 
! particle position and velocity during flight.
  allocate(position_init(3,inpar%Npart))
  allocate(velocity_init(3,inpar%Npart))
  allocate(position(3,inpar%Nsteps))
  allocate(velocity(3,inpar%Nsteps))
  allocate(dens(inpar%Nx,inpar%Ny,inpar%Nz)) ! Proportional to accumulated 
                                             ! density
! Parameters of the grid on which the output is saved
  Lx = inpar%xmax - inpar%xmin
  Ly = inpar%ymax - inpar%ymin
  Lz = inpar%zmax - inpar%zmin
! Set grid cell size
  dxyz(1) = Lx/(inpar%Nx-1)
  dxyz(2) = Ly/(inpar%Ny-1)
  dxyz(3) = Lz/(inpar%Nz-1)

! Initialise
  nextno = 1
  dens = 0.0d0
  velocity_init = 0.0d0
  position_init = 0.0d0


  ! Allocate the probes, get their parameters, and allocate their 
  ! respective arrays for distribution functions etc.
  allocate(probe(inpar%Nprobes))
  attempts = 0
  call GetProbes('', inpar%Nprobes, probe, attempts, Nretries)
  probe%dvx = (probe%vxmax-probe%vxmin)/(probe%Nvx-1)
  probe%dvy = (probe%vymax-probe%vymin)/(probe%Nvy-1)
  probe%dvz = (probe%vzmax-probe%vzmin)/(probe%Nvz-1)
  do ii = 1, inpar%Nprobes
     allocate( probe(ii)%f(probe(ii)%Nvx,probe(ii)%Nvy,probe(ii)%Nvz) )
     probe(ii)%f = 0.0d0
     allocate( probe(ii)%Vx(probe(ii)%Nvx) )
     do jj = 1, probe(ii)%Nvx
        probe(ii)%Vx(jj) = probe(ii)%vxmin + &
             probe(ii)%dvx*dble(jj-1)
     end do
     allocate( probe(ii)%Vy(probe(ii)%Nvy) )
     do jj = 1, probe(ii)%Nvy
        probe(ii)%Vy(jj) = probe(ii)%vymin + &
             probe(ii)%dvy*dble(jj-1)
     end do
     allocate( probe(ii)%Vz(probe(ii)%Nvz) )
     do jj = 1, probe(ii)%Nvz
        probe(ii)%Vz(jj) = probe(ii)%vzmin + &
             probe(ii)%dvz*dble(jj-1)
     end do
  end do

! If we are starting from a dump file, load it now, otherwise 
! proceed with the initialisation of the particles
  if (inpar%startfromdumpfile) then
     attempts = 0
     call LoadParticles(inpar%Npart, position_init, velocity_init, nextno, &
          attempts, Nretries)
     if (nextno>1) then
        attempts = 0
        call LoadDensityDump(inpar%Nx, inpar%Ny, inpar%Nz, dens, &
             nextno, attempts, Nretries)
        attempts = 0
        call LoadProbeDump(inpar%Nprobes, probe, attempts, Nretries)
     end if
  else
     ! Give us particles! 
     ! This is done differently depending on the initialiser that is chosen 
     ! in the file inputcheese.m.
     select case (inpar%initialiser)
     case(1)
        call InitialiseParticles_01(velocity_init,position_init, &
             inpar%Xcube, inpar%Ycube, inpar%Zcube, inpar%vxbulk, & 
             inpar%vybulk, inpar%vzbulk, inpar%vth, inpar%Lside, inpar%Npart)

     case(2)
        call InitialiseParticles_02(velocity_init, position_init, &
             inpar%Npart, inpar%mass, inpar%scdatafilename, &
             inpar%Ngridlevels, grid)

     case(3)
        call InitialiseParticles_03(velocity_init, position_init, &
             inpar%thetamin, inpar%thetamax, inpar%azmin, inpar%azmax, &
             inpar%vrbulk, inpar%vth, inpar%rsource, inpar%Npart)

     case default
        write (*,*) 'initialiser =', inpar%initialiser, ' is illegal!'
        stop
     end select
     attempts = 0
     call DumpParticles(inpar%Npart, position_init, velocity_init, nextno, &
          attempts, Nretries)
     attempts = 0
     call LoadParticles(inpar%Npart, position_init, velocity_init, nextno, &
          attempts, Nretries)
   end if


  
! Here starts the main loop that computes the trajectories
  do ii = nextno, inpar%Npart

! Initial values for the present particle
     position(:,1) = position_init(:,ii)
     velocity(:,1) = velocity_init(:,ii)

! Find initial fields through interpolation
     call FindFields(inpar%Ngridlevels, grid, &
          position(1,1), position(2,1), position(3,1), E0, B0)

! Compute perpendicular E, i.e. remove artificial parallel component
     call PerpendicularEfield(E0, B0, Eperp)

! Find fluctuation E field if we are including such
     if (inpar%Efluctuations) then
        ! To each particle a random phase
        call random_number(rndph)
        fluctphase = 2*pi*rndph
        call FindFluct(inpar%rinfluct, inpar%routfluct, &
             inpar%E1x, inpar%E1y, inpar%E1z, inpar%ffluct,  &
             position(1,1), position(2,1), position(3,1), &
             fluctphase, 0.0d0, E1)
        E1mag = sqrt(sum(E1**2.0d0))
        if ( (inpar%usealpha) .and. (E1mag > 0.0d0) ) then
           call PerpendicularEfield(E1, B0, E1perp)
           E1par = E1 - E1perp
           E1 = E1mag * ( &
                sin(inpar%alpha)*E1par/sqrt(sum(E1par**2.0d0)) + &
                cos(inpar%alpha)*E1perp/sqrt(sum(E1perp**2.0d0)) )
        end if
     end if

! Take half a step back to be ready for leapfrog action
     if (inpar%includephotonpressure) then
        call DustyBoris(velocity(:,1), velocity(:,1), Eperp+E1, B0, &
             -dt/2.0d0, inpar%mass, inpar%charge,inpar%Flight)
     else
        call Boris(velocity(:,1),velocity(:,1),Eperp+E1,B0,-dt/2.0d0, &
             inpar%mass,inpar%charge)
     end if


! Compute the trajectory - start of inner loop
     do jj=1,inpar%Nsteps-1
	 
! Find fields through interpolation
        call FindFields(inpar%Ngridlevels, grid, &
             position(1,jj), position(2,jj), position(3,jj), E0, B0)

! Compute perpendicular E, i.e. remove artificial parallel component
        call PerpendicularEfield(E0, B0, Eperp)

! Find fluctuation E field if we are including such
        if (inpar%Efluctuations) then
           call FindFluct(inpar%rinfluct, inpar%routfluct, &
                inpar%E1x, inpar%E1y, inpar%E1z, inpar%ffluct, &
                position(1,jj), position(2,jj), position(3,jj), &
                fluctphase, dble(jj*dt), E1)
           E1mag = sqrt(sum(E1**2.0d0))
           if ( (inpar%usealpha) .and. (E1mag > 0.0d0) ) then
              call PerpendicularEfield(E1, B0, E1perp)
              E1par = E1 - E1perp
              E1 = E1mag * ( &
                   sin(inpar%alpha)*E1par/sqrt(sum(E1par**2.0d0)) + &
                   cos(inpar%alpha)*E1perp/sqrt(sum(E1perp**2.0d0)) )
           end if
        end if

! Boris gives us the new velocity
        if (inpar%includephotonpressure) then
           call DustyBoris(velocity(:,jj), velocity(:,jj+1), Eperp + E1, B0, &
                dt, inpar%mass, inpar%charge,inpar%Flight)
        else
           call Boris(velocity(:,jj), velocity(:,jj+1), Eperp + E1, B0, dt, &
                inpar%mass, inpar%charge)
        end if

! Now, push!
        position(:,jj+1) = position(:,jj) + velocity(:,jj+1)*dt

! Check that we are still there
        rr=sqrt(sum(position(:,jj+1)**2))
        if ( rr<inpar%robst .or. &
             position(1,jj+1)<inpar%xmin .or. position(1,jj+1)>inpar%xmax .or. &
             position(2,jj+1)<inpar%ymin .or. position(2,jj+1)>inpar%ymax .or. &
             position(3,jj+1)<inpar%zmin .or. position(3,jj+1)>inpar%zmax ) then
           exit
        end if

     end do

     ! Now that the trajectory is computed, accumulate densities to the grid 
     ! and distributions to the distribution function probes, if any.
     call AccumulateDensity(inpar%Nx, inpar%Ny,inpar%Nz, dens, dxyz, &
          inpar%xmin, inpar%ymin, inpar%zmin, inpar%Nsteps, position, &
          jj, dt)
     if ( inpar%Nprobes >0 )then
        call AccumulatefProbe(inpar%Nprobes, probe, &
             inpar%Nsteps, position, velocity, jj, dt)
     end if

     ! Perform periodical dumps periodically
     if ( modulo(ii,inpar%dump_period_dump)==0 ) then
        nextno = ii + 1
        attempts = 0
        call DumpDensityDump(inpar%Nx, inpar%Ny, inpar%Nz, dens, &
             nextno, attempts, Nretries)
        attempts = 0
        call DumpProbeDump(inpar%Nprobes, probe, attempts, Nretries)
        attempts = 0
        call DumpScalarField(dens,inpar%Nx,inpar%Ny,inpar%Nz, &
             'outp/datfiles/statistics/', 'dens', ii, attempts, Nretries)
        do kk = 1, inpar%Nprobes
           write(probeno,fmt='(i2.2)') kk
           probefile = 'fp'//probeno//'N'
           attempts = 0
           call DumpScalarField(probe(kk)%f, &
                probe(kk)%Nvx, probe(kk)%Nvy, probe(kk)%Nvz, &
                'outp/datfiles/statistics/',probefile, ii, attempts, Nretries)
        end do
        write (*,fmt='(a,i10,a)') 'computed ', ii, ' trajectories'
     end if

  end do

! Always dump at the end
  nextno = ii
  attempts = 0
  call DumpDensityDump(inpar%Nx, inpar%Ny, inpar%Nz, dens, &
       nextno, attempts, Nretries)
  attempts = 0
  call DumpProbeDump(inpar%Nprobes, probe, attempts, Nretries)
  attempts = 0
  call DumpScalarField(dens,inpar%Nx,inpar%Ny,inpar%Nz, &
       'outp/datfiles/statistics/', 'dens', ii-1, attempts, Nretries)
  do kk = 1, inpar%Nprobes
     write(probeno,fmt='(i2.2)') kk
     probefile = 'fp'//probeno//'N'
     attempts = 0
     call DumpScalarField(probe(kk)%f, &
          probe(kk)%Nvx, probe(kk)%Nvy, probe(kk)%Nvz, &
          'outp/datfiles/statistics/',probefile, ii-1, attempts, Nretries)
  end do
  write (*,fmt='(a,i10,a)') 'Ending after ', ii-1, ' trajectories'

end program cheddar

